#ifndef _SPADFS_LINUX_COMPAT_H
#define _SPADFS_LINUX_COMPAT_H

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
static inline int blkdev_issue_flush(struct block_device *bdev,
				     sector_t *error_sector)
{
	return 0;
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
static inline void vprintk(const char *fmt, va_list args)
{
	char buffer[256];
	vsnprintf(buffer, sizeof buffer, fmt, args);
	printk("%s", buffer);
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
#define rcu_barrier()	synchronize_rcu()
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
static inline void *kmalloc_node(size_t size, int flags, int node)
{
	return kmalloc(size, flags);
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
static inline void cancel_rearming_delayed_workqueue(
			struct workqueue_struct *wq, struct work_struct *work)
{
	while (!cancel_delayed_work(work))
		flush_workqueue(wq);
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
#define raw_smp_processor_id()	smp_processor_id()
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
static inline int rwsem_is_locked(struct rw_semaphore *sem)
{
	if (down_write_trylock(sem)) {
		up_write(sem);
		return 0;
	}
	return 1;
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
#define mutex			semaphore
#define mutex_init		init_MUTEX
#define mutex_lock		down
#define mutex_unlock		up
#define mutex_trylock(sem)	(!down_trylock(sem))
#define i_mutex			i_sem
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
#define NEW_WORKQUEUE
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
#define PTR_ALIGN(p, a)		((typeof(p))ALIGN((unsigned long)(p), (a)))
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27)
#define get_current_uid()	(current->uid)
#define get_current_gid()	(current->gid)
#define get_current_fsuid()	(current->fsuid)
#define get_current_fsgid()	(current->fsgid)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,5,0)
#define get_current_uid()	current_uid()
#define get_current_gid()	current_gid()
#define get_current_fsuid()	current_fsuid()
#define get_current_fsgid()	current_fsgid()
#else
#define get_current_uid()	from_kuid(&init_user_ns, current_uid())
#define get_current_gid()	from_kgid(&init_user_ns, current_gid())
#define get_current_fsuid()	from_kuid(&init_user_ns, current_fsuid())
#define get_current_fsgid()	from_kgid(&init_user_ns, current_fsgid())
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,5,0)
#define i_uid_read(i)		((i)->i_uid)
#define i_gid_read(i)		((i)->i_gid)
#define i_uid_write(i, u)	((i)->i_uid = (u))
#define i_gid_write(i, u)	((i)->i_gid = (u))
#define uid_eq(a, b)		((a) == (b))
#define gid_eq(a, b)		((a) == (b))
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,9,0)
#define file_inode(f)	((f)->f_dentry->d_inode)
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,19,0)
#define f_dentry	f_path.dentry
#endif

#endif
